/****************************************************************************
** Serial communications unit                                              **
**  by Steven H Don                                                        **
*****************************************************************************
** Unit for communicating with another computer via modem or serial cable. **
**                                                                         **
** For questions, feel free to e-mail me.                                  **
**                                                                         **
**    shd@earthling.net                                                    **
**    http://shd.cjb.net                                                   **
**                                                                         **
****************************************************************************/
#include <string.h>
#include <alloc.h>
#include <dos.h>
#include <values.h>

/*
  Functions available to the calling program are:
    COMOpenPort     : Opens a COM port for read/write access
    COMClosePort    : Closes the COM port that was opened with the above command.
    COMReadByte     : Reads a byte from the buffer, if available
                      Returns 1 if successful.
    COMWriteByte    : Writes a byte to the COM port.
                      Returns 1 if successful.
    COMWriteString  : Writes a string of bytes to the COM port.
                      Returns 1 if successful.
    COMCarrier      : Returns 1 if carrier detect.
*/

void COMOpenPort (char p, char i, long b);
void COMClosePort ();
char COMReadByte (char *b);
char COMWriteByte (char b);
char COMWriteString (char *t);
char COMCarrier ();

//Local variables
  //This stores the address of the old IRQ handler
  #ifdef __cplusplus
    void interrupt (*OldVec)(...);
  #else
    void interrupt (*OldVec)();
  #endif
  //This stores data on the port currently open
  unsigned int       CAddr, CIRQ;
  //This is used to store the incoming data
  char                   *Buffer;
  unsigned int ReadPtr, WritePtr;
  //UART registers must be saved
  char              UARTRegs [2];

/*
  Local functions and procedures:
    COMAddress : Returns the port address of a COM port.
    IRQHandler : Handles the reception of incoming data.
*/

//This returns the hardware port address of the COM port, as reported
//by the BIOS
unsigned int COMAddress (char p)
{
  return peek(0x0040, (p - 1) << 1);
}

//This handles incoming data and stores it in the buffer
#ifdef __cplusplus
  void interrupt IRQHandler (...)
#else
  void interrupt IRQHandler ()
#endif
{
  //Enable interrupts
  asm sti;
  //Store byte in buffer
  Buffer [WritePtr++] = inportb (CAddr);
  //Wrap around buffer
  WritePtr &= 1023;
  //Return interrupts to disable
  asm cli;
  //Acknowledge hardware interrupt
  outportb (0x20, 0x20);
}

/*This opens a COM port for read/write access, it:
- initializes the buffer
- installs the new IRQ handler
- stores the UART registers and reprograms the UART*/
void COMOpenPort (char p, char i, long b)
{
  long BaudRateDivider;

  //Assign buffer
  Buffer = (char *)malloc (1024);
  //Reset read/write pointers
  ReadPtr = 0; WritePtr = 0;

  //Retrieve address
  CAddr = COMAddress (p);

  //Store UART registers
  UARTRegs [0] = inportb (CAddr + 1);
  UARTRegs [1] = inportb (CAddr + 4);

  //Disable interrupts
  asm cli;
  CIRQ = i;
  //Get old handler
  OldVec = getvect (CIRQ + 8);
  //Save new handler
  setvect (CIRQ + 8, IRQHandler);
  //Set UART registers
  outportb (CAddr + 4, 8);
  outportb (CAddr + 1, 1);
  //Enable IRQ
  outportb (0x21, inportb (0x21) & !(1 << CIRQ));
  //Enable interrupts
  asm sti;

  //Initialize the UART
  //set baudrate
  BaudRateDivider = 115200 / b;
  outportb (CAddr + 3, 0x80);
  outportb (CAddr    , BaudRateDivider & 0xFF);
  outportb (CAddr + 1, (BaudRateDivider >> 8) & 0xFF);
  //set to 1 stopbit, 8 databits, no parity
  outportb (CAddr + 3, 3);

  //Relieve the modem
  inportb (CAddr);
  //Reset line status register
  inportb (CAddr + 5);
  //Enable DTR = Data Terminal Ready}
  outportb (CAddr + 4, inportb (CAddr + 4) | 1);
}

//This closes the COM port, undoing COMOpenPort
void COMClosePort ()
{
  //Disable interrupts
  asm cli;
  //Mask IRQ
  outportb (0x21, inportb (0x21) | (1 << CIRQ));
  //Restore UART registers
  outportb (CAddr + 1, UARTRegs [0]);
  outportb (CAddr + 4, UARTRegs [1]);
  //Restore old handler
  setvect (CIRQ + 8, OldVec);
  //Enable interrupts
  asm sti;
}

/*This checks whether a byte is available.
If a byte was found, it returns 1 and "b" contains the value of that byte

If no byte was found, it returns 0.
WARNING: The values returned by "b" in this case have no meaning*/
char COMReadByte (char *b)
{
  //If there is a byte waiting in the buffer, return 1 and store it
  if (ReadPtr != WritePtr) {
    *b = Buffer [ReadPtr++];
    ReadPtr &= 1023;
    return 1;
  } else return 0;
}

/*This writes a byte to the COM port. It returns 1 if successful, 0 if
the port didn't accept the write*/
char COMWriteByte (char b)
{
  unsigned int TimeOut;

  //Wait until the last byte was sent
  do ; while (inportb (CAddr + 5) & 0x20 == 0);
  //Wait for COM port to indicate CTS = Clear To Send, stopping if the
  //connection has been dropped
  TimeOut = MAXINT;
  outportb (CAddr + 4, 0xB);
  do TimeOut--; while (inportb (CAddr + 6) & 0x10 == 0 && TimeOut > 0);
  //If there was no timeout, send the character, otherwise, report failure
  if (TimeOut == 0) return 0;
  outportb (CAddr, b);
  return 1;
}

//This sends an entire string of characters to the port using the above
//procedure
char COMWriteString (char *t)
{
  char Position;

  for (Position = 1; Position <= strlen (t); Position++)
    if (COMWriteByte (t [Position]) == 0) return 0;
  //Successful
  return 1;
}

//This returns 1 if a carrier is detected by the COM port
char COMCarrier ()
{
  if (inportb (CAddr + 6) & 0x80 == 0x80) return 1; else return 0;
}